@@ -5,10 +5,10 @@ import (
 	"encoding/binary"
 	"math/rand/v2"
 	"testing"
-	"time"
 
 	"github.com/quic-go/quic-go/internal/ackhandler"
 	"github.com/quic-go/quic-go/internal/flowcontrol"
+	"github.com/quic-go/quic-go/internal/monotime"
 	"github.com/quic-go/quic-go/internal/protocol"
 	"github.com/quic-go/quic-go/internal/wire"
 
@@ -29,7 +29,7 @@ func TestFramerControlFrames(t *testing.T) {
 		[]ackhandler.Frame{{Frame: &wire.PingFrame{}}},
 		nil,
 		protocol.MaxByteCount,
-		time.Now(),
+		monotime.Now(),
 		protocol.Version1,
 	)
 	require.Len(t, frames, 3)
@@ -51,11 +51,11 @@ func TestFramerControlFrameSizing(t *testing.T) {
 	for i := 0; i < numFrames+1; i++ {
 		framer.QueueControlFrame(bf)
 	}
-	frames, _, length := framer.Append(nil, nil, maxSize, time.Now(), protocol.Version1)
+	frames, _, length := framer.Append(nil, nil, maxSize, monotime.Now(), protocol.Version1)
 	require.Len(t, frames, numFrames)
 	require.Greater(t, length, maxSize-bfLen)
 	// now make sure that the last frame is also added
-	frames, _, length = framer.Append(nil, nil, maxSize, time.Now(), protocol.Version1)
+	frames, _, length = framer.Append(nil, nil, maxSize, monotime.Now(), protocol.Version1)
 	require.Len(t, frames, 1)
 	require.Equal(t, length, bfLen)
 }
@@ -70,7 +70,7 @@ func TestFramerStreamControlFrames(t *testing.T) {
 	framer.QueueControlFrame(ping)
 	str := NewMockStreamControlFrameGetter(gomock.NewController(t))
 	framer.AddStreamWithControlFrames(streamID, str)
-	now := time.Now()
+	now := monotime.Now()
 	str.EXPECT().getControlFrame(now).Return(ackhandler.Frame{Frame: mdf1}, true, true)
 	str.EXPECT().getControlFrame(now).Return(ackhandler.Frame{Frame: mdf2}, true, false)
 	frames, streamFrames, l := framer.Append(nil, nil, protocol.MaxByteCount, now, protocol.Version1)
@@ -91,7 +91,7 @@ func TestFramerStreamControlFramesSizing(t *testing.T) {
 	framer := newFramer(flowcontrol.NewConnectionFlowController(0, 0, nil, nil, nil))
 	framer.AddStreamWithControlFrames(10, str)
 	str.EXPECT().getControlFrame(gomock.Any()).Return(ackhandler.Frame{Frame: mdf1}, true, true).AnyTimes()
-	frames, _, l := framer.Append(nil, nil, 100, time.Now(), protocol.Version1)
+	frames, _, l := framer.Append(nil, nil, 100, monotime.Now(), protocol.Version1)
 	require.Equal(t, protocol.ByteCount(len(frames))*mdf1.Length(protocol.Version1), l)
 	require.Greater(t, l, protocol.ByteCount(100-maxStreamControlFrameSize))
 	require.LessOrEqual(t, l, protocol.ByteCount(100))
@@ -135,7 +135,7 @@ func testFramerStreamDataBlocked(t *testing.T, fits bool) {
 	)
 
 	const maxSize protocol.ByteCount = 1000
-	frames, streamFrames, l := framer.Append(nil, nil, maxSize, time.Now(), protocol.Version1)
+	frames, streamFrames, l := framer.Append(nil, nil, maxSize, monotime.Now(), protocol.Version1)
 	require.Len(t, streamFrames, 1)
 	dataLen := streamFrames[0].Frame.DataLen()
 	if fits {
@@ -144,7 +144,7 @@ func testFramerStreamDataBlocked(t *testing.T, fits bool) {
 	} else {
 		require.Equal(t, streamFrames[0].Frame.Length(protocol.Version1), l)
 		require.Empty(t, frames)
-		frames, streamFrames, l2 := framer.Append(nil, nil, maxSize, time.Now(), protocol.Version1)
+		frames, streamFrames, l2 := framer.Append(nil, nil, maxSize, monotime.Now(), protocol.Version1)
 		require.Greater(t, l+l2, maxSize)
 		require.Empty(t, streamFrames)
 		require.Len(t, frames, 1)
@@ -193,7 +193,7 @@ func testFramerDataBlocked(t *testing.T, fits bool) {
 	)
 
 	const maxSize protocol.ByteCount = 1000
-	frames, streamFrames, l := framer.Append(nil, nil, maxSize, time.Now(), protocol.Version1)
+	frames, streamFrames, l := framer.Append(nil, nil, maxSize, monotime.Now(), protocol.Version1)
 	require.Len(t, streamFrames, 1)
 	if fits {
 		require.Len(t, frames, 1)
@@ -201,7 +201,7 @@ func testFramerDataBlocked(t *testing.T, fits bool) {
 	} else {
 		require.Equal(t, streamFrames[0].Frame.Length(protocol.Version1), l)
 		require.Empty(t, frames)
-		frames, streamFrames, l2 := framer.Append(nil, nil, maxSize, time.Now(), protocol.Version1)
+		frames, streamFrames, l2 := framer.Append(nil, nil, maxSize, monotime.Now(), protocol.Version1)
 		require.Greater(t, l+l2, maxSize)
 		require.Empty(t, streamFrames)
 		require.Len(t, frames, 1)
@@ -215,7 +215,7 @@ func TestFramerDetectsFrameDoS(t *testing.T) {
 		framer.QueueControlFrame(&wire.PingFrame{})
 		framer.QueueControlFrame(&wire.PingFrame{})
 		require.False(t, framer.QueuedTooManyControlFrames())
-		frames, _, _ := framer.Append([]ackhandler.Frame{}, nil, 1, time.Now(), protocol.Version1)
+		frames, _, _ := framer.Append([]ackhandler.Frame{}, nil, 1, monotime.Now(), protocol.Version1)
 		require.Len(t, frames, 1)
 		require.Len(t, framer.controlFrames, i+1)
 	}
@@ -238,13 +238,13 @@ func TestFramerDetectsFramePathResponseDoS(t *testing.T) {
 	}
 	for i := 0; i < maxPathResponses; i++ {
 		require.True(t, framer.HasData())
-		frames, _, length := framer.Append(nil, nil, protocol.MaxByteCount, time.Now(), protocol.Version1)
+		frames, _, length := framer.Append(nil, nil, protocol.MaxByteCount, monotime.Now(), protocol.Version1)
 		require.Len(t, frames, 1)
 		require.Equal(t, pathResponses[i], frames[0].Frame)
 		require.Equal(t, pathResponses[i].Length(protocol.Version1), length)
 	}
 	require.False(t, framer.HasData())
-	frames, _, length := framer.Append(nil, nil, protocol.MaxByteCount, time.Now(), protocol.Version1)
+	frames, _, length := framer.Append(nil, nil, protocol.MaxByteCount, monotime.Now(), protocol.Version1)
 	require.Empty(t, frames)
 	require.Zero(t, length)
 }
@@ -260,14 +260,14 @@ func TestFramerPacksSinglePathResponsePerPacket(t *testing.T) {
 	framer.QueueControlFrame(cf1)
 	framer.QueueControlFrame(cf2)
 	// the first packet should contain a single PATH_RESPONSE frame, but all the other control frames
-	frames, _, _ := framer.Append(nil, nil, protocol.MaxByteCount, time.Now(), protocol.Version1)
+	frames, _, _ := framer.Append(nil, nil, protocol.MaxByteCount, monotime.Now(), protocol.Version1)
 	require.Len(t, frames, 3)
 	require.Equal(t, f1, frames[0].Frame)
 	require.Contains(t, []wire.Frame{frames[1].Frame, frames[2].Frame}, cf1)
 	require.Contains(t, []wire.Frame{frames[1].Frame, frames[2].Frame}, cf2)
 	// the second packet should contain the other PATH_RESPONSE frame
 	require.True(t, framer.HasData())
-	frames, _, _ = framer.Append(nil, nil, protocol.MaxByteCount, time.Now(), protocol.Version1)
+	frames, _, _ = framer.Append(nil, nil, protocol.MaxByteCount, monotime.Now(), protocol.Version1)
 	require.Len(t, frames, 1)
 	require.Equal(t, f2, frames[0].Frame)
 	require.False(t, framer.HasData())
@@ -285,7 +285,7 @@ func TestFramerAppendStreamFrames(t *testing.T) {
 	framer := newFramer(flowcontrol.NewConnectionFlowController(0, 0, nil, nil, nil))
 	require.False(t, framer.HasData())
 	// no frames added yet
-	controlFrames, fs, length := framer.Append(nil, nil, protocol.MaxByteCount, time.Now(), protocol.Version1)
+	controlFrames, fs, length := framer.Append(nil, nil, protocol.MaxByteCount, monotime.Now(), protocol.Version1)
 	require.Empty(t, controlFrames)
 	require.Empty(t, fs)
 	require.Zero(t, length)
@@ -304,7 +304,7 @@ func TestFramerAppendStreamFrames(t *testing.T) {
 	// Even though the first stream claimed to have more data,
 	// we only dequeue a single STREAM frame per call of AppendStreamFrames.
 	f0 := ackhandler.StreamFrame{Frame: &wire.StreamFrame{StreamID: 9999}}
-	controlFrames, fs, length = framer.Append([]ackhandler.Frame{}, []ackhandler.StreamFrame{f0}, protocol.MaxByteCount, time.Now(), protocol.Version1)
+	controlFrames, fs, length = framer.Append([]ackhandler.Frame{}, []ackhandler.StreamFrame{f0}, protocol.MaxByteCount, monotime.Now(), protocol.Version1)
 	require.Empty(t, controlFrames)
 	require.Len(t, fs, 3)
 	require.Equal(t, f0, fs[0])
@@ -322,7 +322,7 @@ func TestFramerAppendStreamFrames(t *testing.T) {
 
 	// ... but it actually doesn't
 	str1.EXPECT().popStreamFrame(gomock.Any(), protocol.Version1).Return(ackhandler.StreamFrame{}, nil, false)
-	_, fs, length = framer.Append(nil, nil, protocol.MaxByteCount, time.Now(), protocol.Version1)
+	_, fs, length = framer.Append(nil, nil, protocol.MaxByteCount, monotime.Now(), protocol.Version1)
 	require.Empty(t, fs)
 	require.Zero(t, length)
 	require.False(t, framer.HasData())
@@ -336,7 +336,7 @@ func TestFramerRemoveActiveStream(t *testing.T) {
 	require.True(t, framer.HasData())
 	framer.RemoveActiveStream(id) // no calls will be issued to the mock stream
 	// we can't assert on framer.HasData here, since it's not removed from the ringbuffer
-	_, frames, _ := framer.Append(nil, nil, protocol.MaxByteCount, time.Now(), protocol.Version1)
+	_, frames, _ := framer.Append(nil, nil, protocol.MaxByteCount, monotime.Now(), protocol.Version1)
 	require.Empty(t, frames)
 	require.False(t, framer.HasData())
 }
@@ -349,7 +349,7 @@ func TestFramerMinStreamFrameSize(t *testing.T) {
 
 	require.True(t, framer.HasData())
 	// don't pop frames smaller than the minimum STREAM frame size
-	_, frames, _ := framer.Append(nil, nil, protocol.MinStreamFrameSize-1, time.Now(), protocol.Version1)
+	_, frames, _ := framer.Append(nil, nil, protocol.MinStreamFrameSize-1, monotime.Now(), protocol.Version1)
 	require.Empty(t, frames)
 
 	// pop frames of the minimum size
@@ -360,7 +360,7 @@ func TestFramerMinStreamFrameSize(t *testing.T) {
 			return ackhandler.StreamFrame{Frame: f}, nil, false
 		},
 	)
-	_, frames, _ = framer.Append(nil, nil, protocol.MinStreamFrameSize, time.Now(), protocol.Version1)
+	_, frames, _ = framer.Append(nil, nil, protocol.MinStreamFrameSize, monotime.Now(), protocol.Version1)
 	require.Len(t, frames, 1)
 	// unsetting DataLenPresent on the last frame reduced the size slightly beyond the minimum size
 	require.Equal(t, protocol.MinStreamFrameSize-2, frames[0].Frame.Length(protocol.Version1))
@@ -380,7 +380,7 @@ func TestFramerMinStreamFrameSizeMultipleStreamFrames(t *testing.T) {
 	}
 	str.EXPECT().popStreamFrame(gomock.Any(), protocol.Version1).Return(ackhandler.StreamFrame{Frame: f}, nil, false)
 	framer.AddActiveStream(id, str)
-	_, fs, length := framer.Append(nil, nil, 500, time.Now(), protocol.Version1)
+	_, fs, length := framer.Append(nil, nil, 500, monotime.Now(), protocol.Version1)
 	require.Len(t, fs, 1)
 	require.Equal(t, f, fs[0].Frame)
 	require.Equal(t, f.Length(protocol.Version1), length)
@@ -404,7 +404,7 @@ func TestFramerFillPacketOneStream(t *testing.T) {
 			},
 		)
 		framer.AddActiveStream(id, str)
-		_, frames, _ := framer.Append(nil, nil, i, time.Now(), protocol.Version1)
+		_, frames, _ := framer.Append(nil, nil, i, monotime.Now(), protocol.Version1)
 		require.Len(t, frames, 1)
 		require.False(t, frames[0].Frame.DataLenPresent)
 		// make sure the entire space was filled up
@@ -440,7 +440,7 @@ func TestFramerFillPacketMultipleStreams(t *testing.T) {
 		)
 		framer.AddActiveStream(id1, stream1)
 		framer.AddActiveStream(id2, stream2)
-		_, frames, _ := framer.Append(nil, nil, i, time.Now(), protocol.Version1)
+		_, frames, _ := framer.Append(nil, nil, i, monotime.Now(), protocol.Version1)
 		require.Len(t, frames, 2)
 		require.True(t, frames[0].Frame.DataLenPresent)
 		require.False(t, frames[1].Frame.DataLenPresent)
@@ -467,7 +467,7 @@ func TestFramer0RTTRejection(t *testing.T) {
 	framer.AddActiveStream(10, NewMockStreamFrameGetter(gomock.NewController(t)))
 
 	framer.Handle0RTTRejection()
-	controlFrames, streamFrames, _ := framer.Append(nil, nil, protocol.MaxByteCount, time.Now(), protocol.Version1)
+	controlFrames, streamFrames, _ := framer.Append(nil, nil, protocol.MaxByteCount, monotime.Now(), protocol.Version1)
 	require.Empty(t, streamFrames)
 	require.Len(t, controlFrames, 3)
 	require.Contains(t, controlFrames, ackhandler.Frame{Frame: pc})
