@@ -1,11 +1,11 @@
 package main
 
 import (
+	"crypto/rand"
 	"log"
+	mrand "math/rand/v2"
 	"time"
 
-	"golang.org/x/exp/rand"
-
 	"github.com/quic-go/quic-go"
 	"github.com/quic-go/quic-go/fuzzing/internal/helper"
 	"github.com/quic-go/quic-go/internal/protocol"
@@ -21,15 +21,15 @@ func getRandomData(l int) []byte {
 }
 
 func getRandomNumber() uint64 {
-	switch 1 << uint8(rand.Intn(3)) {
+	switch 1 << uint8(mrand.IntN(3)) {
 	case 1:
-		return uint64(rand.Int63n(64))
+		return mrand.Uint64N(64)
 	case 2:
-		return uint64(rand.Int63n(16384))
+		return mrand.Uint64N(16384)
 	case 4:
-		return uint64(rand.Int63n(1073741824))
+		return mrand.Uint64N(1073741824)
 	case 8:
-		return uint64(rand.Int63n(4611686018427387904))
+		return mrand.Uint64N(4611686018427387904)
 	default:
 		panic("unexpected length")
 	}
@@ -39,15 +39,14 @@ func getRandomNumberLowerOrEqual(target uint64) uint64 {
 	if target == 0 {
 		return 0
 	}
-	return uint64(rand.Int63n(int64(target)))
+	return mrand.Uint64N(target)
 }
 
 // returns a *maximum* number of num ACK ranges
 func getAckRanges(num int) []wire.AckRange {
-	var ranges []wire.AckRange
-
-	prevSmallest := uint64(rand.Int63n(4611686018427387904))
-	for i := 0; i < num; i++ {
+	prevSmallest := mrand.Uint64N(4611686018427387904)
+	ranges := make([]wire.AckRange, 0, num)
+	for range num {
 		if prevSmallest <= 2 {
 			break
 		}
@@ -262,21 +261,21 @@ func main() {
 		}
 	}
 
-	for i := 0; i < 30; i++ {
+	for range 30 {
 		frames := getFrames()
 
 		var b []byte
-		for j := 0; j < rand.Intn(30)+2; j++ {
-			if rand.Intn(10) == 0 { // write a PADDING frame
+		for range mrand.IntN(30) + 2 {
+			if mrand.IntN(10) == 0 { // write a PADDING frame
 				b = append(b, 0)
 			}
-			f := frames[rand.Intn(len(frames))]
+			f := frames[mrand.IntN(len(frames))]
 			var err error
 			b, err = f.Append(b, version)
 			if err != nil {
 				log.Fatal(err)
 			}
-			if rand.Intn(10) == 0 { // write a PADDING frame
+			if mrand.IntN(10) == 0 { // write a PADDING frame
 				b = append(b, 0)
 			}
 		}
