@@ -15,9 +15,8 @@ import (
 )
 
 func TestKeyUpdates(t *testing.T) {
-	origKeyUpdateInterval := handshake.KeyUpdateInterval
-	t.Cleanup(func() { handshake.KeyUpdateInterval = origKeyUpdateInterval })
-	handshake.KeyUpdateInterval = 1 // update keys as frequently as possible
+	reset := handshake.SetKeyUpdateInterval(1) // update keys as frequently as possible
+	t.Cleanup(reset)
 
 	var sentHeaders []*logging.ShortHeader
 	var receivedHeaders []*logging.ShortHeader
@@ -40,7 +39,7 @@ func TestKeyUpdates(t *testing.T) {
 		return
 	}
 
-	server, err := quic.Listen(newUPDConnLocalhost(t), getTLSConfig(), nil)
+	server, err := quic.Listen(newUDPConnLocalhost(t), getTLSConfig(), nil)
 	require.NoError(t, err)
 	defer server.Close()
 
@@ -48,7 +47,7 @@ func TestKeyUpdates(t *testing.T) {
 	defer cancel()
 	conn, err := quic.Dial(
 		ctx,
-		newUPDConnLocalhost(t),
+		newUDPConnLocalhost(t),
 		server.Addr(),
 		getTLSClientConfig(),
 		getQuicConfig(&quic.Config{Tracer: func(context.Context, logging.Perspective, quic.ConnectionID) *logging.ConnectionTracer {
