@@ -14,7 +14,7 @@ import (
 )
 
 func TestStatelessResets(t *testing.T) {
-	t.Run("0 byte connection IDs", func(t *testing.T) {
+	t.Run("zero-length connection IDs", func(t *testing.T) {
 		testStatelessReset(t, 0)
 	})
 	t.Run("10 byte connection IDs", func(t *testing.T) {
@@ -26,11 +26,10 @@ func testStatelessReset(t *testing.T, connIDLen int) {
 	var statelessResetKey quic.StatelessResetKey
 	rand.Read(statelessResetKey[:])
 
-	c := newUPDConnLocalhost(t)
+	c := newUDPConnLocalhost(t)
 	tr := &quic.Transport{
-		Conn:               c,
-		StatelessResetKey:  &statelessResetKey,
-		ConnectionIDLength: connIDLen,
+		Conn:              c,
+		StatelessResetKey: &statelessResetKey,
 	}
 	defer tr.Close()
 	ln, err := tr.Listen(getTLSConfig(), getQuicConfig(nil))
@@ -58,25 +57,38 @@ func testStatelessReset(t *testing.T, connIDLen int) {
 
 	var drop atomic.Bool
 	proxy := quicproxy.Proxy{
-		Conn:       newUPDConnLocalhost(t),
+		Conn:       newUDPConnLocalhost(t),
 		ServerAddr: ln.Addr().(*net.UDPAddr),
-		DropPacket: func(quicproxy.Direction, []byte) bool { return drop.Load() },
+		DropPacket: func(quicproxy.Direction, net.Addr, net.Addr, []byte) bool { return drop.Load() },
 	}
 	require.NoError(t, proxy.Start())
 	defer proxy.Close()
 
-	cl := &quic.Transport{
-		Conn:               newUPDConnLocalhost(t),
-		ConnectionIDLength: connIDLen,
+	var conn *quic.Conn
+	if connIDLen > 0 {
+		cl := &quic.Transport{
+			Conn:               newUDPConnLocalhost(t),
+			ConnectionIDLength: connIDLen,
+		}
+		defer cl.Close()
+		var err error
+		conn, err = cl.Dial(
+			context.Background(),
+			proxy.LocalAddr(),
+			getTLSClientConfig(),
+			getQuicConfig(&quic.Config{MaxIdleTimeout: 2 * time.Second}),
+		)
+		require.NoError(t, err)
+	} else {
+		conn, err = quic.Dial(
+			context.Background(),
+			newUDPConnLocalhost(t),
+			proxy.LocalAddr(),
+			getTLSClientConfig(),
+			getQuicConfig(&quic.Config{MaxIdleTimeout: 2 * time.Second}),
+		)
+		require.NoError(t, err)
 	}
-	defer cl.Close()
-	conn, err := cl.Dial(
-		context.Background(),
-		proxy.LocalAddr(),
-		getTLSClientConfig(),
-		getQuicConfig(&quic.Config{MaxIdleTimeout: 2 * time.Second}),
-	)
-	require.NoError(t, err)
 	str, err := conn.AcceptStream(context.Background())
 	require.NoError(t, err)
 	data := make([]byte, 6)
@@ -94,9 +106,8 @@ func testStatelessReset(t *testing.T, connIDLen int) {
 	// We need to create a new Transport here, since the old one is still sending out
 	// CONNECTION_CLOSE packets for (recently) closed connections).
 	tr2 := &quic.Transport{
-		Conn:               c,
-		ConnectionIDLength: connIDLen,
-		StatelessResetKey:  &statelessResetKey,
+		Conn:              c,
+		StatelessResetKey: &statelessResetKey,
 	}
 	defer tr2.Close()
 	ln2, err := tr2.Listen(getTLSConfig(), getQuicConfig(nil))
