@@ -11,11 +11,12 @@ import (
 
 	"github.com/quic-go/quic-go/internal/protocol"
 	"github.com/quic-go/quic-go/internal/qerr"
+	"github.com/quic-go/quic-go/quicvarint"
 	"github.com/stretchr/testify/require"
 )
 
 func TestFrameTypeParsingReturnsNilWhenNothingToRead(t *testing.T) {
-	parser := NewFrameParser(true, true)
+	parser := NewFrameParser(true, true, true)
 	frameType, l, err := parser.ParseType(nil, protocol.Encryption1RTT)
 	require.Equal(t, io.EOF, err)
 	require.Zero(t, frameType)
@@ -23,7 +24,7 @@ func TestFrameTypeParsingReturnsNilWhenNothingToRead(t *testing.T) {
 }
 
 func TestParseLessCommonFrameReturnsEOFWhenNothingToRead(t *testing.T) {
-	parser := NewFrameParser(true, true)
+	parser := NewFrameParser(true, true, true)
 	l, f, err := parser.ParseLessCommonFrame(FrameTypeMaxStreamData, nil, protocol.Version1)
 	require.IsType(t, &qerr.TransportError{}, err)
 	require.Zero(t, l)
@@ -31,7 +32,7 @@ func TestParseLessCommonFrameReturnsEOFWhenNothingToRead(t *testing.T) {
 }
 
 func TestFrameParsingSkipsPaddingFrames(t *testing.T) {
-	parser := NewFrameParser(true, true)
+	parser := NewFrameParser(true, true, true)
 	b := []byte{0, 0} // 2 PADDING frames
 	b, err := (&PingFrame{}).Append(b, protocol.Version1)
 	require.NoError(t, err)
@@ -48,7 +49,7 @@ func TestFrameParsingSkipsPaddingFrames(t *testing.T) {
 }
 
 func TestFrameParsingHandlesPaddingAtEnd(t *testing.T) {
-	parser := NewFrameParser(true, true)
+	parser := NewFrameParser(true, true, true)
 	b := []byte{0, 0, 0}
 
 	_, l, err := parser.ParseType(b, protocol.Encryption1RTT)
@@ -57,7 +58,7 @@ func TestFrameParsingHandlesPaddingAtEnd(t *testing.T) {
 }
 
 func TestFrameParsingParsesSingleFrame(t *testing.T) {
-	parser := NewFrameParser(true, true)
+	parser := NewFrameParser(true, true, true)
 	var b []byte
 	for range 10 {
 		var err error
@@ -76,7 +77,7 @@ func TestFrameParsingParsesSingleFrame(t *testing.T) {
 }
 
 func TestFrameParserACK(t *testing.T) {
-	parser := NewFrameParser(true, true)
+	parser := NewFrameParser(true, true, true)
 	f := &AckFrame{AckRanges: []AckRange{{Smallest: 1, Largest: 0x13}}}
 	b, err := f.Append(nil, protocol.Version1)
 	require.NoError(t, err)
@@ -102,7 +103,7 @@ func TestFrameParserAckDelay(t *testing.T) {
 }
 
 func testFrameParserAckDelay(t *testing.T, encLevel protocol.EncryptionLevel) {
-	parser := NewFrameParser(true, true)
+	parser := NewFrameParser(true, true, true)
 	parser.SetAckDelayExponent(protocol.AckDelayExponent + 2)
 	f := &AckFrame{
 		AckRanges: []AckRange{{Smallest: 1, Largest: 1}},
@@ -136,7 +137,7 @@ func checkFrameUnsupported(t *testing.T, err error, expectedFrameType uint64) {
 }
 
 func TestFrameParserStreamFrames(t *testing.T) {
-	parser := NewFrameParser(true, true)
+	parser := NewFrameParser(true, true, true)
 	f := &StreamFrame{
 		StreamID: 0x42,
 		Offset:   0x1337,
@@ -159,7 +160,7 @@ func TestFrameParserStreamFrames(t *testing.T) {
 }
 
 func TestParseStreamFrameWrapsError(t *testing.T) {
-	parser := NewFrameParser(true, true)
+	parser := NewFrameParser(true, true, true)
 	f := &StreamFrame{
 		StreamID:       0x1234,
 		Offset:         0x1000,
@@ -187,7 +188,7 @@ func TestParseStreamFrameWrapsError(t *testing.T) {
 }
 
 func TestParseStreamFrameSuccess(t *testing.T) {
-	parser := NewFrameParser(true, true)
+	parser := NewFrameParser(true, true, true)
 	original := &StreamFrame{
 		StreamID:       0x1234,
 		Offset:         0x1000,
@@ -319,23 +320,38 @@ func TestFrameParserFrames(t *testing.T) {
 			frameType: FrameTypeResetStreamAt,
 			frame:     &ResetStreamFrame{StreamID: 0x1337, ReliableSize: 0x42, FinalSize: 0xdeadbeef},
 		},
+		{
+			name:      "ACK_FREQUENCY",
+			frameType: FrameTypeAckFrequency,
+			frame: &AckFrequencyFrame{
+				SequenceNumber:        0x1337,
+				AckElicitingThreshold: 0x42,
+				RequestMaxAckDelay:    123 * time.Second,
+				ReorderingThreshold:   0xcafe,
+			},
+		},
+		{
+			name:      "IMMEDIATE_ACK",
+			frameType: FrameTypeImmediateAck,
+			frame:     &ImmediateAckFrame{},
+		},
 	}
 
 	for _, test := range tests {
 		t.Run(test.name, func(t *testing.T) {
-			parser := NewFrameParser(true, true)
+			parser := NewFrameParser(true, true, true)
 			b, err := test.frame.Append(nil, protocol.Version1)
 			require.NoError(t, err)
 
 			frameType, l, err := parser.ParseType(b, protocol.Encryption1RTT)
 			require.NoError(t, err)
 			require.Equal(t, test.frameType, frameType)
-			require.Equal(t, 1, l)
+			require.Equal(t, quicvarint.Len(uint64(test.frameType)), l)
 
 			frame, l, err := parser.ParseLessCommonFrame(frameType, b[l:], protocol.Version1)
 			require.NoError(t, err)
 			require.Equal(t, test.frame, frame)
-			require.Equal(t, len(b)-1, l)
+			require.Equal(t, len(b)-quicvarint.Len(uint64(test.frameType)), l)
 		})
 	}
 }
@@ -453,7 +469,7 @@ func TestFrameAllowedAtEncLevel(t *testing.T) {
 					allowed = tc.allowedOneRTT
 				}
 
-				parser := NewFrameParser(true, true)
+				parser := NewFrameParser(true, true, true)
 				b, err := tc.frame.Append(nil, protocol.Version1)
 				require.NoError(t, err)
 				frameType, _, err := parser.ParseType(b, encLevel)
@@ -472,7 +488,7 @@ func TestFrameAllowedAtEncLevel(t *testing.T) {
 }
 
 func TestFrameParserDatagramFrame(t *testing.T) {
-	parser := NewFrameParser(true, true)
+	parser := NewFrameParser(true, true, true)
 	f := &DatagramFrame{
 		Data: []byte("foobar"),
 	}
@@ -496,7 +512,7 @@ func TestFrameParserDatagramFrame(t *testing.T) {
 }
 
 func TestFrameParserDatagramUnsupported(t *testing.T) {
-	parser := NewFrameParser(false, true)
+	parser := NewFrameParser(false, true, true)
 	f := &DatagramFrame{Data: []byte("foobar")}
 	b, err := f.Append(nil, protocol.Version1)
 	require.NoError(t, err)
@@ -506,17 +522,42 @@ func TestFrameParserDatagramUnsupported(t *testing.T) {
 }
 
 func TestFrameParserResetStreamAtUnsupported(t *testing.T) {
-	parser := NewFrameParser(true, false)
+	parser := NewFrameParser(true, false, true)
 	f := &ResetStreamFrame{StreamID: 0x1337, ReliableSize: 0x42, FinalSize: 0xdeadbeef}
 	b, err := f.Append(nil, protocol.Version1)
 	require.NoError(t, err)
 
 	_, _, err = parser.ParseType(b, protocol.Encryption1RTT)
-	checkFrameUnsupported(t, err, 0x24)
+	checkFrameUnsupported(t, err, uint64(FrameTypeResetStreamAt))
+}
+
+func TestFrameParserAckFrequencyUnsupported(t *testing.T) {
+	parser := NewFrameParser(true, true, false)
+
+	t.Run("ACK_FREQUENCY", func(t *testing.T) {
+		f := &AckFrequencyFrame{
+			SequenceNumber:        1337,
+			AckElicitingThreshold: 42,
+			RequestMaxAckDelay:    42 * time.Millisecond,
+			ReorderingThreshold:   1234,
+		}
+		b, err := f.Append(nil, protocol.Version1)
+		require.NoError(t, err)
+		_, _, err = parser.ParseType(b, protocol.Encryption1RTT)
+		checkFrameUnsupported(t, err, uint64(FrameTypeAckFrequency))
+	})
+
+	t.Run("IMMEDIATE_ACK", func(t *testing.T) {
+		f := &ImmediateAckFrame{}
+		b, err := f.Append(nil, protocol.Version1)
+		require.NoError(t, err)
+		_, _, err = parser.ParseType(b, protocol.Encryption1RTT)
+		checkFrameUnsupported(t, err, uint64(FrameTypeImmediateAck))
+	})
 }
 
 func TestFrameParserInvalidFrameType(t *testing.T) {
-	parser := NewFrameParser(true, true)
+	parser := NewFrameParser(true, true, true)
 
 	_, l, err := parser.ParseType(encodeVarInt(0x42), protocol.Encryption1RTT)
 
@@ -529,7 +570,7 @@ func TestFrameParserInvalidFrameType(t *testing.T) {
 }
 
 func TestFrameParsingErrorsOnInvalidFrames(t *testing.T) {
-	parser := NewFrameParser(true, true)
+	parser := NewFrameParser(true, true, true)
 	f := &MaxStreamDataFrame{
 		StreamID:          0x1337,
 		MaximumStreamData: 0xdeadbeef,
@@ -711,7 +752,7 @@ func TestFrameParserAllocs(t *testing.T) {
 
 func testFrameParserAllocs(t *testing.T, frames []Frame) float64 {
 	buf := writeFrames(t, frames...)
-	parser := NewFrameParser(true, true)
+	parser := NewFrameParser(true, true, true)
 	parser.SetAckDelayExponent(3)
 
 	return testing.AllocsPerRun(100, func() {
@@ -777,15 +818,13 @@ func BenchmarkParseDatagramFrame(b *testing.B) {
 }
 
 func benchmarkFrames(b *testing.B, frames ...Frame) {
-	buf := writeFrames(b, frames...)
-
-	parser := NewFrameParser(true, true)
-	parser.SetAckDelayExponent(3)
-
-	b.ResetTimer()
 	b.ReportAllocs()
 
-	for range b.N {
+	buf := writeFrames(b, frames...)
+	parser := NewFrameParser(true, true, true)
+	parser.SetAckDelayExponent(3)
+
+	for b.Loop() {
 		parseFrames(b, parser, buf, frames...)
 	}
 }
