@@ -13,12 +13,13 @@ type MaxStreamsFrame struct {
 	MaxStreamNum protocol.StreamNum
 }
 
-func parseMaxStreamsFrame(b []byte, typ uint64, _ protocol.Version) (*MaxStreamsFrame, int, error) {
+func parseMaxStreamsFrame(b []byte, typ FrameType, _ protocol.Version) (*MaxStreamsFrame, int, error) {
 	f := &MaxStreamsFrame{}
+	//nolint:exhaustive // Function will only be called with BidiMaxStreamsFrameType or UniMaxStreamsFrameType
 	switch typ {
-	case bidiMaxStreamsFrameType:
+	case FrameTypeBidiMaxStreams:
 		f.Type = protocol.StreamTypeBidi
-	case uniMaxStreamsFrameType:
+	case FrameTypeUniMaxStreams:
 		f.Type = protocol.StreamTypeUni
 	}
 	streamID, l, err := quicvarint.Parse(b)
@@ -35,9 +36,9 @@ func parseMaxStreamsFrame(b []byte, typ uint64, _ protocol.Version) (*MaxStreams
 func (f *MaxStreamsFrame) Append(b []byte, _ protocol.Version) ([]byte, error) {
 	switch f.Type {
 	case protocol.StreamTypeBidi:
-		b = append(b, bidiMaxStreamsFrameType)
+		b = append(b, byte(FrameTypeBidiMaxStreams))
 	case protocol.StreamTypeUni:
-		b = append(b, uniMaxStreamsFrameType)
+		b = append(b, byte(FrameTypeUniMaxStreams))
 	}
 	b = quicvarint.Append(b, uint64(f.MaxStreamNum))
 	return b, nil
